'use strict';

var require$$0 = require('fs');
var path = require('path');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var require$$0__default = /*#__PURE__*/_interopDefaultLegacy(require$$0);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);

const {copyFileSync: copyFileSync$1, existsSync, readFileSync: readFileSync$1} = require$$0__default['default'];


/**
 *
 * @param {any} parameter
 * @returns {boolean} true if param is a string
 */
function isString(parameter) {
	return (typeof parameter === 'string' || parameter instanceof String);
}

/**
 *
 * @param {{
 * 	hostingSite: string|undefined;
 * 	sourceRewriteMatch: string;
 * 	firebaseJson: string
 * }} param
 * @returns {{
 * 	functions: false | { name: string, source: string };
 * 	cloudRun: false | { serviceId: string, region: string };
 * 	publicDir: string
 * }}
 */
function parseFirebaseConfiguration$1({hostingSite, sourceRewriteMatch, firebaseJson}) {
	if (!existsSync(firebaseJson)) {
		throw new Error(`File ${firebaseJson} does not exist. The provided file should exist and be a Firebase JSON config.`);
	}

	let firebaseConfig;
	try {
		firebaseConfig = JSON.parse(readFileSync$1(firebaseJson, 'utf-8'));
	} catch (error) {
		throw new Error(`Error parsing ${firebaseJson}. ${error.message}`);
	}

	if (!firebaseConfig?.hosting) {
		throw new TypeError(`Error with config ${firebaseJson}. "hosting" field required.`);
	}

	// Force "hosting" field to be array
	const firebaseHostingConfig = Array.isArray(firebaseConfig.hosting) ? firebaseConfig.hosting : [{...firebaseConfig.hosting}];

	// Force "site" field to be included in "hosting" if more than 1 hosting site config
	if (firebaseHostingConfig.length > 1) {
		for (const item of firebaseHostingConfig) {
			if (!item.site) {
				throw new Error(`Error with config ${firebaseJson}. "hosting" configs should identify their "site" name as Firebase supports multiple sites. This site config does not ${JSON.stringify(item)}`);
			}
		}
	}

	const hostingConfig = firebaseHostingConfig.length === 1 ?
		firebaseHostingConfig[0] :
		firebaseHostingConfig.find(item => item.site === hostingSite);

	if (!hostingConfig) {
		throw new Error(`Error with config ${firebaseJson}. No "hosting[].site" match for ${hostingSite}. Ensure your svelte.config.js adapter config "hostingSite" matches an item in your Firebase config.`);
	}

	if (!isString(hostingConfig?.public)) {
		throw new Error(`Error with config ${firebaseJson}. "hosting[].public" field is required and should be a string. Hosting config with error: ${JSON.stringify(hostingConfig)}`);
	}

	if (!Array.isArray(hostingConfig?.rewrites)) {
		throw new TypeError(`Error with config ${firebaseJson}. "hosting[].rewrites" field  required in hosting config and should be an array of object(s). Hosting config with error: ${JSON.stringify(hostingConfig)}`);
	}

	const rewriteConfig = hostingConfig.rewrites.find(item => {
		return item.source === sourceRewriteMatch && (item.function || item.run);
	});

	if (!rewriteConfig) {
		throw new Error(`Error with config ${firebaseJson}. "hosting[].rewrites[*]" does not contain a config with "source":"${sourceRewriteMatch}" and either "function" or "run". Is your "sourceRewriteMatch" in svelte.config.js correct?`);
	}

	if (rewriteConfig?.run && (!rewriteConfig.run.serviceId || !isString(rewriteConfig.run.serviceId))) {
		throw new Error(`Error with config ${firebaseJson}. Cloud Run rewrite configuration missing required field "serviceId". Rewrite config with error: ${JSON.stringify(rewriteConfig)}`);
	}

	if (rewriteConfig?.run && !validCloudRunServiceId(rewriteConfig.run.serviceId)) {
		throw new Error(`Error with config ${firebaseJson}. The "serviceId":"${rewriteConfig.run.serviceId}" must use only lowercase alphanumeric characters and dashes, cannot begin or end with a dash, and cannot be longer than 63 characters.`);
	}

	if (rewriteConfig?.run && rewriteConfig?.run?.region && rewriteConfig.run.region !== 'us-central1') {
		throw new Error(`Error with config ${firebaseJson}. Firebase Hosting rewrites only support "regions":"us-central1" (docs - https://firebase.google.com/docs/functions/locations#http_and_client-callable_functions). Change "${rewriteConfig.run.region}" accordingly.`);
	}

	if (rewriteConfig?.function && !validCloudFunctionName(rewriteConfig.function)) {
		throw new Error(`Error with config ${firebaseJson}. The "serviceId":"${rewriteConfig.function}" must use only alphanumeric characters and underscores and cannot be longer than 62 characters.`);
	}

	// If function, ensure function root-level field is present
	if (rewriteConfig?.function && (!firebaseConfig?.functions || !firebaseConfig.functions?.source || !isString(firebaseConfig.functions.source))) {
		throw new Error(`Error with config ${firebaseJson}. If you're using Cloud Functions for your SSR rewrite rule, you need to define a "functions.source" field (of type string) at your config root.`);
	}

	return {
		functions: rewriteConfig?.function ? {
			name: rewriteConfig.function,
			source: firebaseConfig.functions.source
		} : false,
		cloudRun: rewriteConfig?.run ? {
			serviceId: rewriteConfig.run.serviceId,
			region: rewriteConfig.run?.region || 'us-central1'
		} : false,
		publicDir: hostingConfig.public
	};
}

/**
 * Cloud Run Service ID rules:
 * - only lowercase alphanumeric characters and dashes
 * - cannot begin or end with a dash
 * - cannot be longer than 63 characters
 * @param {string} serviceId
 * @returns {boolean} `true` if valid
 */
function validCloudRunServiceId(serviceId) {
	return /^[a-z\d][a-z\d-]+[a-z\d]$/gm.test(serviceId) && serviceId.length < 64;
}

/**
 * Cloud Function name rules:
 * - alphanumeric
 * - underscore
 * - max length 62 chars
 * @param {string} name
 * @returns {boolean} `true` if valid
 *
 * Rules a combination of
 * - https://github.com/firebase/firebase-tools/blob/1633f4fccbbc1bcbc6216fe13b8e888c8940bde4/src/deploy/functions/validate.ts#L38
 * - https://github.com/firebase/firebase-tools/blob/2dc7216a498dee2ca7e2acc33d6ba16d5647e27f/src/extractTriggers.js#L18
 */
function validCloudFunctionName(name) {
	return /^\w{1,62}$/.test(name);
}

/**
 * Copy File If Exists (synchronously)
 * @param {string} filename source file path
 * @param {string} destDir destination directory
 */
function copyFileIfExistsSync$1(filename, destDir) {
	if (existsSync(filename)) {
		copyFileSync$1(filename, path__default['default'].join(destDir, filename));
	}
}

var utils = {
	parseFirebaseConfiguration: parseFirebaseConfiguration$1,
	validCloudRunServiceId,
	validCloudFunctionName,
	copyFileIfExistsSync: copyFileIfExistsSync$1
};

const {copyFileSync, readFileSync, renameSync, writeFileSync} = require$$0__default['default'];

const {copyFileIfExistsSync, parseFirebaseConfiguration} = utils;

/**
 * @param {{
 * 	hostingSite?: string;
 * 	sourceRewriteMatch?: string,
 * 	firebaseJson?: string,
 * 	cloudRunBuildDir?: string
 * }} options
 */
var cli = function ({
	firebaseJson = 'firebase.json',
	hostingSite = undefined,
	sourceRewriteMatch = '**',
	cloudRunBuildDir = undefined
} = {}) {
	/** @type {import('@sveltejs/kit').Adapter} */
	const adapter = {
		name: 'svelte-adapter-firebase',
		async adapt(builder) {
			const {functions, cloudRun, publicDir} = parseFirebaseConfiguration({hostingSite, sourceRewriteMatch, firebaseJson});

			if (functions !== false) {
				adaptToCloudFunctions({builder, ...functions});
			}

			if (cloudRun !== false) {
				adaptToCloudRun({builder, ...cloudRun, cloudRunBuildDir});
			}

			builder.log.info(`Writing client application to ${publicDir}`);
			builder.copy_static_files(publicDir);
			builder.copy_client_files(publicDir);

			builder.log.info(`Prerendering static pages to ${publicDir}`);
			await builder.prerender({dest: publicDir});
		}
	};

	return adapter;
};

/**
 *
 * @param {{
 * 	builder: any;
 * 	name: string;
 * 	source: string;
 * }} param
 */
function adaptToCloudFunctions({builder, name, source}) {
	const functionsPackageJson = JSON.parse(readFileSync(path__default['default'].join(source, 'package.json'), 'utf-8'));
	const functionsMain = functionsPackageJson?.main;
	if (!functionsMain) {
		throw new Error(`Error reading ${functionsPackageJson}. Required field "main" missing.`);
	}

	const ssrDirname = name ?? 'svelteKit';
	const serverOutputDir = path__default['default'].join(source, path__default['default'].dirname(functionsMain), ssrDirname);

	builder.log.minor(`Writing Cloud Function server assets to ${serverOutputDir}`);
	builder.copy_server_files(serverOutputDir);

	// Prepare handler & entrypoint
	renameSync(path__default['default'].join(serverOutputDir, 'app.js'), path__default['default'].join(serverOutputDir, 'app.mjs'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'index.js'), path__default['default'].join(serverOutputDir, 'index.js'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'handler.mjs'), path__default['default'].join(serverOutputDir, 'handler.mjs'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'handler.mjs.map'), path__default['default'].join(serverOutputDir, 'handler.mjs.map'));

	// Prepare Cloud Function
	const functionsEntrypoint = path__default['default'].join(source, functionsMain);
	try {
		const ssrSvelteFunctionName = ssrDirname.replace(/\W/g, '').concat('Server');
		if (!readFileSync(functionsEntrypoint, 'utf-8').includes(`${name} =`)) {
			builder.log.warn(
				// eslint-disable-next-line indent
`Add the following Cloud Function to ${functionsEntrypoint}
+--------------------------------------------------+
let ${ssrSvelteFunctionName};
exports.${name} = functions.https.onRequest(async (request, response) => {
	if (!${ssrSvelteFunctionName}) {
		functions.logger.info("Initializing SvelteKit SSR Handler");
		${ssrSvelteFunctionName} = require("./${ssrDirname}/index").default;
		functions.logger.info("SvelteKit SSR Handler initialised!");
	}
	return await ${ssrSvelteFunctionName}(request, response);
});
+--------------------------------------------------+`
			);
		}
	} catch (error) {
		throw new Error(`Error reading Cloud Function entrypoint file: ${functionsEntrypoint}. ${error.message}`);
	}
}

/**
 *
 * @param {{
 * 	builder: any;
 * 	serviceId: string;
 * 	region: string;
 * 	cloudRunBuildDir: string|undefined
 * }} param
 */
function adaptToCloudRun({builder, serviceId, region, cloudRunBuildDir}) {
	const serverOutputDir = path__default['default'].join(cloudRunBuildDir || `.${serviceId}`);

	builder.log.info(`Writing Cloud Run service to ./${serverOutputDir}`);
	builder.copy_server_files(serverOutputDir);

	// Prepare handler & entrypoint
	renameSync(path__default['default'].join(serverOutputDir, 'app.js'), path__default['default'].join(serverOutputDir, 'app.mjs'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'index.js'), path__default['default'].join(serverOutputDir, 'index.js'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'handler.mjs'), path__default['default'].join(serverOutputDir, 'handler.mjs'));
	copyFileSync(path__default['default'].join(__dirname, 'files', 'handler.mjs.map'), path__default['default'].join(serverOutputDir, 'handler.mjs.map'));

	// Prepare Cloud Run package.json - read SvelteKit App 'package.json', modify the JSON, write to serverOutputDir
	const pkgjson = JSON.parse(readFileSync('package.json', 'utf-8'));
	pkgjson.scripts.start = 'functions-framework --target=default';
	pkgjson.dependencies['@google-cloud/functions-framework'] = '^1.7.1'; // Peer-dep of this adapter instead?
	pkgjson.engines = {node: '14'};
	delete pkgjson.type;
	const data = JSON.stringify(pkgjson, null, 2);
	writeFileSync(path__default['default'].join(serverOutputDir, 'package.json'), data);

	// Copy lockfile for deps install during image build
	copyFileIfExistsSync('package-lock.json', serverOutputDir);
	copyFileIfExistsSync('npm-shrinkwrap.json', serverOutputDir);
	copyFileIfExistsSync('yarn.lock', serverOutputDir);
	copyFileIfExistsSync('pnpm-lock.yaml', serverOutputDir);

	builder.log.warn(
		// eslint-disable-next-line indent
`To deploy your Cloud Run service, run this command:
+--------------------------------------------------+
gcloud beta run deploy ${serviceId} --platform managed --region ${region} --source ${serverOutputDir} --allow-unauthenticated
+--------------------------------------------------+`
	);
}

module.exports = cli;
